% lab10_ex_resampling_2_pp.m
%
% Introduction to POLYPHASE DECOMPOSITION: doing resampling faster!
%
% Polyphase interpolation, sampling K-times UP.
% Polyphase decimation, sampling L-times DOWN.
%
% In QUICK INTRODUCTTION it was shown that K-times up-sampling is implemented as
% a low-pass filtering of original signal with artificially inserted K-1 zeros
% between every two signal samples.
% There is no sense to multiply the interpolation filter weights by inserted zeros.
% Therefore ONLY every K-th filter weight is used for computing an output sample.
%
% POLYPHASE IMPLEMENTATION OF INTERPOLATION FILTER
% Let assume that the interpolation filter has 6 weights h(n) (K*M=6, K=3, M=2)
% and the zero-inserted signal xz(n) has some values.
% We will calculate interpolated samples xi(n) for n=7, 8, 9, 10, 11, 12:
%
%         k  =  1,  2,  3,  4,  5,  6,  7,  8,  9,  10,  11,  12,  13,  14,  15,  16,
%         n  =  1,  2,  3,  4,  5,  6,  7,  8,  9,  10,  11,  12,  13,  14,  15,  16,
%      xz(n) = 10,  0,  0, 20,  0,  0, 30,  0,  0,  40,   0,   0,  50,   0,   0,  60, ...    
%     h(7-k) =  0,  6,  5,  4,  3,  2,  1,  0,  0,   0,   0,   0,   0,   0,   0,   0, ...
%     h(8-k) =  0,  0,  6,  5,  4,  3,  2,  1,  0,   0,   0,   0,   0,   0,   0,   0, ...
%     h(9-k) =  0,  0,  0,  6,  5,  4,  3,  2,  1,   0,   0,   0,   0,   0,   0,   0, ...
%    h(10-k) =  0,  0,  0,  0,  6,  5,  4,  3,  2,   1,   0,   0,   0,   0,   0,   0, ...
%    h(11-k) =  0,  0,  0,  0,  0,  6,  5,  4,  3,   2,   1,   0,   0,   0,   0,   0, ...
%    h(12-k) =  0,  0,  0,  0,  0,  0,  6,  5,  4,   3,   2,   1,   0,   0,   0,   0, ...
%               h  xz    h  xz
%      xi(7) = (1)*30 + (4)*20 =      110                          
%      xi(8) = (2)*30 + (5)*20 =          160                          
%      xi(9) = (3)*30 + (6)*20 =              210                          
%     xi(10) = (1)*40 + (4)*30 =                   160                          
%     xi(11) = (2)*40 + (5)*30 =                        230                          
%     xi(12) = (3)*40 + (6)*30 =                             300
%
% What do we see?
% That zero insertion can be avoided.
% We should create a so-called polyphase sequences of h(n):
%     h1(n) = [ 1, 4 ]  - start from h(1) and take every K-th filter weight
%     h2(n) = [ 2, 5 ]  - start from h(2) and take every K-th filter weight
%     h3(n) = [ 3, 6 ]  - start from h(3) and take every K-th filter weight
% and then apply them one-by-one in merry-go-round fashion to original signal weighting.
% This way multiplications by eliminated.
%
% In turn, in L-order down-sampling procedure we are applying a low-pass filter
% to the original signal samples and leave only one sample from L calculated ones.
% What's the point of calculation of something and not using it? None!
% 
% POLYPHASE IMPLEMENTATION OF DECIMATION FILTER
% In this case we are simply shifting the filter weights NO ONE POSITION FORWARD
% but L positions FORWARD (L=3).
%        x(n)      =  ..., 1,  2,  3,  4,  5,  6,  7,  8,  9,  10,  11,  12, ...
%        h(n0-k)   =  ..., 0,  6,  5,  4,  3,  2,  1,  0,  0,   0,   0,   0,  ...
%       h(n0+L-k)  =  ..., 0,  0,  0,  0,  6,  5,  4,  3,  2,   1,   0,   0,  ...
%       h(n0+2L-k) =  ..., 0,  0,  0,  0,  0,  0,  0,  6,  5,   4,   3,   2,  ...

clear all; close all;

% #########################################################################
% INTERPOLATION ###########################################################
% #########################################################################

K = 5;                      % required UP-SAMPLING (interpolation) order

% Signal
N = 200;                    % number of input samples
x = sin(2*pi/20*(0:N-1));   % input signal
ni = 1 : 400;               % for error/display

% Design of LowPass filter coefficients - using window method
M = 200; Nh = 2*M+1;
h = fir1(2*M+1,1/K);
Nz = rem(Nh,K); Napp=K-Nz; h(Nh+1:Nh+Napp) = zeros(1,Napp); Nh=Nh+Napp;  % appending zeros at the end of h(n)

% INTERPOLATION: Inefficient implementation - multiplication by zeros
Ni = K*N;                          % number of samples after interpolation
xz = zeros(1,Ni); xz(1:K:Ni) = x;  % insertion of zeros between signal samples
xi = filter(h,1,xz);               % low-pass filtering - smoothing

% INTERPOLATION: Fast polyphase implementation - multiplication by zeros are removed
xipp = zeros(1,length(x));                      % output initialization                         
for k=1:K                                       % LOOP START: PP filtering
    xipp(k:K:Ni) = filter( h(k:K:end), 1, x );  % k-th PP component filtering
end                                             % LOOP END
MaxERR_INTERPpp = max(abs(xi-xipp)),            % error
figure; plot(ni,xi(ni),'ro-',ni,xipp(ni),'bx-'); grid; xlabel('n');
title('Interpolated signals - fast PP'); pause

% INTERPOLATION: Slow illustrative implementation by PP sections with inserted zeros 
figure;
xipps = zeros(1,Ni);                            % output initialization 
for k=1:K                                       % LOOP START: PP filtering
    hk = zeros(1,length(h)); hk(k:K:end) = h(k:K:end);  % PP component with zeros
    work = filter( hk, 1, xz );                         % k-th PP component filtering
    xipps = xipps + work;                       % accumulation of filtering resuls
  % xipps(k:K:Ni) = work(k:K:Ni);               % the same as above 
    subplot(211); stem(work(1:40)); hold on;    % figure of k-th PP filtering result
    subplot(212); stem(xipps(1:40)); pause      % figure of accumulated result 
end                                             % LOOP END
MaxERR_INTERPpps = max(abs(xi-xipps)),          % error
figure; plot(ni,xi(ni),'ro-',ni,xipps(ni),'bx-'); grid; xlabel('n');
title('Interpolated signals - slow PP'); pause

% #########################################################################
% DECIMATION ##############################################################
% #########################################################################

L = 4;                      % required DOWN-SAMPLING (decimation) order

% Signal
N = 500*L;                  % number of input samples
x = sin(2*pi/40*(0:N-1));   % input signal
nd = 1 : 100;               % for error/display

% Design of LowPass filter coefficients - using window method
M = 200; Nh = 2*M+1;
h = fir1(2*M+1,1/L-0.01);
Nz = rem(Nh,L); Napp=L-Nz; h(Nh+1:Nh+Napp) = zeros(1,Napp); Nh=Nh+Napp;  % appending zeros at the end of h(n)

% DECTIMATION: Inefficient - output samples are calculated and removed
Nz = rem(N,L); Napp=L-Nz; x(N+1:N+Napp) = zeros(1,Napp);  N=N+Napp; % appending zeros
xd = filter(h,1,x); xd = xd(1:L:end);

% DECIMATION: Fast polyphase implementation - samples to be removed are not computed
xz = [ zeros(1,L-1) x(1:end-(L-1)) ]; % initial insertion of L-1 preceding zeros 
for k=1:L                             % LOOP START: separate PP filtering
    xdppm(k,:) = filter( h(k:L:end), 1, xz(L-k+1:L:end) ); % k-th PP section
end                                   % LOOP END
xdpp = sum(xdppm);                    % summation of PP components
MaxERR_DECIMpp = max(abs(xd-xdpp)),   % error
figure; plot(nd,xd(nd),'ro-',nd,xdpp(nd),'bx-'); xlabel('n');
title('Decimated signals - fast PP'); pause

% DECIMATION: Slow illustrative implementation with PP sections with inserted zeros 
xdpm = zeros( L, length(x) );         % PP matrix initialization
for k=1:L                             % LOOP START: separate PP filtering
    hk = zeros(1,length(h)); hk(k:L:end) = h(k:L:end); % k-th PP filter with zeros 
    xdpm(k,:) = filter( hk, 1, x);    % k-th PP section with inserted zeros
end                                   % LOOP END
xdp = sum( xdpm );                    % summation of PP components
xdp = xdp(1:L:end);                   % decimation
MaxERR_INTERPpdp = max(abs(xd-xdp)),  % error
figure; plot(nd,xd(nd),'ro-',nd,xdp(nd),'bx-'); grid; xlabel('n');
title('Decimated signals - slow PP'); pause
