% lab10_ex_resampling_demo1 - resampling of real-world signals
% UP sampling and DOWN sampling of speech signals using Matlab functions:
% interp(), decimate(), resample()

clear all; close all; subplot(111);

K = 3;   % UP-sampling K-times
L = 2;   % DOWN-sampling L-times

% Reading signal from disc
disp('Speech reading from HD ...');
%[x,fs,Nb ] = wavread('speech.wav'); x=x';   % old Matlab
[x,fs ] = audioread('speech.wav'); x=x';    % new Matlab
%fs=8000; Nsec=3; x = sin(2*pi*100/fs*(0:Nsec*fs-1)) + sin(2*pi*1000/fs*(0:Nsec*fs-1));
Nx=length(x); 
dt=1/fs; t=dt*(0:Nx-1);
figure; plot(t,x); xlabel('t [s]'); title('x(t)'); grid;
soundsc(x,fs); pause

fsUP = K*fs;        % new K-times higher sampling frequency
fsDOWN = fsUP/L;    % new L-times lower  sampling frequency 

% Interpolation via appending zeros in FFT spectrum: M times UP
disp('Speech interpolation via FFT ...'); pause
K = 3;                                   % UP-sampling order
N = 2^14;                                % number of signal samples
Nz = (K-1)*N+1;                          % number of zeros to be added
y = x(1:N);                              % cut signal fragment
Y = fft(y);                              % perform its FFT
Y = [ Y(1:N/2) zeros(1,Nz) Y(N/2+2:N)];  % append zeros in the middle, ensure symmetry
yUP = K*real(ifft(Y));                   % perform IFFT
figure; plot(dt*(0:N-1),y,'r.-',dt/K*(0:K*N-1),yUP,'b.-');
grid; xlabel('t [s]'); title('x(t) and xUP(t) - for FFT WITH ZEROS'); pause
soundsc(yUP,fsUP); pause

% Interpolation (K times UP) via Matlab function
disp('Speech interpolation via Matlab interp() ...');
xUP = interp(x,K);
figure; plot(t,x,'r.-',dt/K*(0:K*Nx-1),xUP,'b.-');
grid; xlabel('t [s]'); title('x(t) and xUP(t) - for INTERP()');
soundsc(xUP,fsUP); pause

% Decimation (L times DOWN) via Matlab function
disp('Speech decimation via Matlab decimate() ...');
xDOWN1 = decimate(xUP,L);
figure; plot(t,x,'r.-',dt/K*(0:K*Nx-1),xUP,'b.-',L*dt/K*(0:K*Nx/L-1)+dt/K,xDOWN1,'g.-');
grid; xlabel('t [s]'); title('x(t),  xUP(t),  xDOWN1(t)');
soundsc(xDOWN1,fsDOWN); pause

% Re-sampling (K times UP and L times DOWN) via Matlab function 
disp('Speech resampling via Matlab resample() ...');
xDOWN2 = resample(x,K,L);
figure; plot(L*dt/K*(0:K*Nx/L-1)+dt/K,xDOWN1,'g.-',L*dt/K*(0:K*Nx/L-1),xDOWN2,'r.-');
grid; xlabel('t [s]'); title('xDOWN1(t),  xDOWN2(t)');
soundsc(xDOWN2,fsDOWN); pause
